@@ -101,7 +101,7 @@ unghostify(cPersistentObject *self)
         {
             /* Create a node in the ring for this unghostified object. */
             self->cache->non_ghost_count++;
-            self->cache->total_estimated_size += 
+            self->cache->total_estimated_size +=
                 _estimated_size_in_bytes(self->estimated_size);
             ring_add(&self->cache->ring_home, &self->ring);
             Py_INCREF(self);
@@ -152,13 +152,14 @@ static void
 ghostify(cPersistentObject *self)
 {
     PyObject **dictptr, *slotnames;
+    PyObject *errtype, *errvalue, *errtb;
 
     /* are we already a ghost? */
     if (self->state == cPersistent_GHOST_STATE)
         return;
 
     /* Is it ever possible to not have a cache? */
-    if (self->cache == NULL) 
+    if (self->cache == NULL)
     {
         self->state = cPersistent_GHOST_STATE;
         return;
@@ -177,7 +178,7 @@ ghostify(cPersistentObject *self)
     /* If we're ghostifying an object, we better have some non-ghosts. */
     assert(self->cache->non_ghost_count > 0);
     self->cache->non_ghost_count--;
-    self->cache->total_estimated_size -= 
+    self->cache->total_estimated_size -=
         _estimated_size_in_bytes(self->estimated_size);
     ring_del(&self->ring);
     self->state = cPersistent_GHOST_STATE;
@@ -195,6 +196,12 @@ ghostify(cPersistentObject *self)
      *   override __new__ ) */
     if (Py_TYPE(self)->tp_new == Pertype.tp_new)
     {
+        /* later we might clear an AttributeError but
+         * if we have a pending exception that still needs to be
+         * raised so that we don't generate a SystemError.
+         */
+        PyErr_Fetch(&errtype, &errvalue, &errtb);
+
         slotnames = pickle_slotnames(Py_TYPE(self));
         if (slotnames && slotnames != Py_None)
         {
@@ -235,6 +242,7 @@ ghostify(cPersistentObject *self)
             }
         }
         Py_XDECREF(slotnames);
+        PyErr_Restore(errtype, errvalue, errtb);
     }
 
     /* We remove the reference to the just ghosted object that the ring
@@ -262,7 +270,7 @@ changed(cPersistentObject *self)
         if (meth == NULL)
             return -1;
         arg = PyTuple_New(1);
-        if (arg == NULL) 
+        if (arg == NULL)
         {
             Py_DECREF(meth);
             return -1;
@@ -371,7 +379,7 @@ pickle_slotnames(PyTypeObject *cls)
             return NULL;
         if (n)
             slotnames = Py_None;
-        
+
         Py_INCREF(slotnames);
         return slotnames;
     }
@@ -594,7 +602,7 @@ pickle___setstate__(PyObject *self, PyObject *state)
         int len;
 
         dict = _PyObject_GetDictPtr(self);
-        
+
         if (!dict)
         {
             PyErr_SetString(PyExc_TypeError,
@@ -775,6 +783,7 @@ Per__getstate__(cPersistentObject *self)
 static void
 Per_dealloc(cPersistentObject *self)
 {
+    PyObject_GC_UnTrack((PyObject *)self);
     if (self->state >= 0)
     {
         /* If the cache has been cleared, then a non-ghost object
@@ -1363,6 +1372,101 @@ Per_set_sticky(cPersistentObject *self, PyObject* value)
     return 0;
 }
 
+static PyObject*
+repr_format_exception(char* format)
+{
+    /* If an exception we should catch occurred, return a new
+       string of its repr. Otherwise, return NULL. */
+    PyObject *exc_t;
+    PyObject *exc_v;
+    PyObject *exc_tb;
+    PyObject *result = NULL;
+
+    if (PyErr_Occurred() && PyErr_ExceptionMatches(PyExc_Exception))
+    {
+        PyErr_Fetch(&exc_t, &exc_v, &exc_tb);
+        PyErr_NormalizeException(&exc_t, &exc_v, &exc_tb);
+        PyErr_Clear();
+
+        result = PyUnicode_FromFormat(format, exc_v);
+        Py_DECREF(exc_t);
+        Py_DECREF(exc_v);
+        Py_DECREF(exc_tb);
+    }
+    return result;
+}
+
+static PyObject*
+repr_helper(PyObject *o, char* format)
+{
+    /* Returns a new reference, or NULL on error */
+    PyObject *result;
+
+    if (o)
+    {
+        result = PyUnicode_FromFormat(format, o);
+        if (!result)
+            result = repr_format_exception(format);
+    }
+    else
+    {
+        result = PyUnicode_FromString("");
+    }
+
+    return result;
+
+}
+
+static PyObject*
+Per_repr(cPersistentObject *self)
+{
+    PyObject *prepr = NULL;
+    PyObject *prepr_exc_str = NULL;
+
+    PyObject *oid_str = NULL;
+    PyObject *jar_str = NULL;
+    PyObject *result = NULL;
+
+    prepr = PyObject_GetAttrString((PyObject*)Py_TYPE(self), "_p_repr");
+    if (prepr)
+    {
+        result = PyObject_CallFunctionObjArgs(prepr, self, NULL);
+        if (result)
+            goto cleanup;
+        else
+        {
+            prepr_exc_str = repr_format_exception(" _p_repr %R");
+            if (!prepr_exc_str)
+                goto cleanup;
+        }
+    }
+    else
+    {
+        PyErr_Clear();
+        prepr_exc_str = PyUnicode_FromString("");
+    }
+
+    oid_str = repr_helper(self->oid, " oid %R");
+    if (!oid_str)
+        goto cleanup;
+
+    jar_str = repr_helper(self->jar, " in %R");
+    if (!jar_str)
+        goto cleanup;
+
+    result = PyUnicode_FromFormat("<%s object at %p%S%S%S>",
+                                  Py_TYPE(self)->tp_name, self,
+                                  oid_str, jar_str, prepr_exc_str);
+
+cleanup:
+    Py_XDECREF(prepr);
+    Py_XDECREF(prepr_exc_str);
+    Py_XDECREF(oid_str);
+    Py_XDECREF(jar_str);
+
+    return result;
+}
+
 static PyGetSetDef Per_getsets[] = {
   {"_p_changed", (getter)Per_get_changed, (setter)Per_set_changed},
   {"_p_jar", (getter)Per_get_jar, (setter)Per_set_jar},
@@ -1445,7 +1549,7 @@ static PyTypeObject Pertype = {
   0,                                /* tp_getattr */
   0,                                /* tp_setattr */
   0,                                /* tp_compare */
-  0,                                /* tp_repr */
+  (reprfunc)Per_repr,               /* tp_repr */
   0,                                /* tp_as_number */
   0,                                /* tp_as_sequence */
   0,                                /* tp_as_mapping */
